﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.IO;

namespace Black.NBT
{
    /// <summary>
    /// Représente un TAG_List
    /// </summary>
    public class TagList : Tag, ITagHierarchicalData, ICloneable
    {
        #region Propriétés

        /// <summary>
        /// Obtient le nombre d'élément
        /// </summary>
        public int Count
        {
            get
            {
                return (this.Items.Count);
            }
        }

        /// <summary>
        /// Obtient la liste des éléments
        /// </summary>
        public ObservableCollection<Tag> Items
        {
            get
            {
                return (this._items);
            }
            private set
            {
                if (this._items != value)
                {
                    this._items = value;
                    this.OnPropertyChanged();
                }
            }
        }

        /// <summary>
        /// Obtient ou définit le type de la liste
        /// </summary>
        public TagType Type
        {
            get
            {
                return (this._tagType);
            }
            set
            {
                if (this._tagType != value)
                {
                    this._tagType = value;
                    this.OnPropertyChanged();
                }
            }
        }

        /// <summary>
        /// Obtient ou définit un élément de la liste
        /// </summary>
        /// <param name="index">Index de l'élément</param>
        /// <returns>Instance de l'index</returns>
        public Tag this[int index]
        {
            get
            {
                if ( index <0 || index >= this.Count)
                    throw(new IndexOutOfRangeException());

                return (this.Items[index]);
            }
            set
            {
                if (index < 0 || index >= this.Count)
                    throw (new IndexOutOfRangeException());

                this.Items[index] = value;
            }
        }

        #endregion

        #region Variables d'instances

        private TagType _tagType;
        private ObservableCollection<Tag> _items;

        #endregion

        #region Constructeur de la classe

        /// <summary>
        /// Constructeur de la classe
        /// </summary>
        /// <param name="name">Nom du tag</param>
        public TagList(string name)
            : base(name)
        {
            this.TagType = NBT.TagType.List;
            this._items = new ObservableCollection<Tag>();
        }

        #endregion

        #region Fonctions internes

        /// <summary>
        /// Ecrit l'instance sur le stream
        /// </summary>
        /// <param name="stream">Instance du flux</param>
        /// <param name="fromList">Si l'écriture vient d'une liste</param>
        internal override void WriteToStream(MemoryStream stream, bool fromList)
        {
            // Si pas d'une liste on inscrit tout 

            if (!fromList)
            {
                // Ecriture du type

                stream.WriteByte((byte)this.TagType);

                // Ecriture du nom


                NBTFile.WriteString(stream, this.Name);
            }

            // Ecriture type de liste

            NBTFile.WriteByte(stream, (byte)this.Type);

            // Ecriture nombre d'élément

            NBTFile.WriteInt(stream, this.Items.Count);

            // Ecriture des éléménts

            foreach (Tag item in this.Items)
                item.WriteToStream(stream, true);
        }

        #endregion

        #region Fonctions statiques

        /// <summary>
        /// Lit le tag dans le flux
        /// </summary>
        /// <param name="stream">Instance du flux placé</param>
        /// <returns>Instance du Tag</returns>
        public static TagList Read(MemoryStream stream)
        {
            return (TagList.Read(stream, true));
        }

        /// <summary>
        /// Lit le tag dans le flux
        /// </summary>
        /// <param name="stream">Instance du flux placé</param>
        /// <param name="withName">TRUE récupère le nom, FALSE sinon</param>
        /// <returns>Instance du Tag</returns>
        public static TagList Read(MemoryStream stream, bool withName)
        {
            List<Tag> list = new List<Tag>();

            // On récupère son nom

            string name = string.Empty;

            if (withName)
                name = Tag.ReadName(stream);

            // Création de l'instance

            TagList item = new TagList(name);

            // Récupération du type de la liste

            item.Type = (TagType)NBTFile.ReadByte(stream);

            // Récupération du nombre d'élément

            int length = NBTFile.ReadInt(stream);

            // On parcours la liste

            for (int i = 0; i < length; i++)
            {
                // Extraction du tag

                switch (item.Type)
                {
                    case (TagType.Byte):
                        list.Add(TagByte.Read(stream, false));
                        break;

                    case (TagType.ByteArray):
                        list.Add(TagByteArray.Read(stream, false));
                        break;

                    case (TagType.Compound):
                        list.Add(TagCompound.Read(stream, false));
                        break;

                    case (TagType.Double):
                        list.Add(TagDouble.Read(stream, false));
                        break;

                    case (TagType.Float):
                        list.Add(TagFloat.Read(stream, false));
                        break;

                    case (TagType.Int):
                        list.Add(TagInt.Read(stream, false));
                        break;

                    case (TagType.List):
                        list.Add(TagList.Read(stream, false));
                        break;

                    case (TagType.Long):
                        list.Add(TagLong.Read(stream, false));
                        break;

                    case (TagType.Short):
                        list.Add(TagShort.Read(stream, false));
                        break;

                    case (TagType.String):
                        list.Add(TagString.Read(stream, false));
                        break;

                    default:
                        throw (new InvalidDataException("Read invalid or unknow Tag Data Type."));
                }
            }

            // On tri les enfants

            list.Sort();

            // Affectation de la liste

            foreach (Tag tag in list)
                item.Items.Add(tag);

            // On retourne la valeur

            return (item);
        }

        #endregion

        #region Clonage

        /// <summary>
        /// Clone l'instance
        /// </summary>
        /// <returns>Nouvelle instance</returns>
        object ICloneable.Clone()
        {
            return (this.Clone());
        }

        /// <summary>
        /// Clone l'instance
        /// </summary>
        /// <returns>Nouvelle instance</returns>
        public virtual TagList Clone()
        {
            TagList clone = this.MemberwiseClone() as TagList;

            clone.Items = new ObservableCollection<Tag>();

            foreach (ICloneable item in this.Items)
                clone.Items.Add((Tag)item.Clone());

            return (clone);
        }

        #endregion
    }
}
